#!/bin/sh
FS=tank/home
ZFSDIR=/Volumes/$FS/dds
export BDIR=/Users/dds/src/transaction/bench

run_ssh_unpack()
{
	cd $DIR
	/usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-ssh-unpack.time tar xzf $BDIR/openssh-4.2p1.tar.gz
	cd $BDIR
}

run_ssh_config()
{
	cd $DIR/openssh-4.2p1
	/usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-ssh-config.time sh -c './configure >&config.out'
	cd $BDIR
}

run_ssh_make()
{
	cd $DIR/openssh-4.2p1
	/usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-ssh-make.time sh -c 'make >&make.out'
	cd $BDIR
}

run_mysql_unpack()
{
	cd $DIR
	mkdir $DIR/local
	echo 'tar xzf $BDIR/mysql-5.1.30.tar.gz' | /usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-mysql-unpack.time su -m dds
	cd $BDIR
}

run_mysql_config()
{
	cd $DIR/mysql-5.1.30
	echo './configure --prefix=$DIR/local >&config.out' | /usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-mysql-config.time su -m dds
	cd $BDIR
}

run_mysql_make()
{
	cd $DIR/mysql-5.1.30
	echo 'make >&make.out' | /usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-mysql-make.time su -m dds
	cd $BDIR
}

run_mysql_test()
{
	cd $DIR/mysql-5.1.30
	echo 'make test-force >&test.out' | /usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-mysql-test.time su -m dds
	cd $BDIR
}

run_mysql_install()
{
	cd $DIR/mysql-5.1.30
	echo 'make install >&install.out' | /usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-mysql-test.time su -m dds
	cd $BDIR
}

run_trashdir()
{
	/usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-trashdir.time $BDIR/trashdir $DIR/openssh-4.2p1
}

run_postmark()
{
	mkdir $DIR/postmark-run
	cd $DIR/postmark-run
	/usr/bin/time $TIME_OPTS 2> $BDIR/$NAME-postmark.time $BDIR/postmark >$BDIR/$NAME-postmark.out <<\EOF
set size 5120 102400
set buffering false
set read 65536
set write 65536
set number 2500
set transactions 9000
run
EOF
	cd $BDIR
}

# Prepare to execute a benchmark set  with the specified name on the specified directory
run_prepare()
{
	NAME=$1
	date | tr -d \\n
	echo ":  starting $NAME"
	wait_zfs
	zfs umount tank/home
	zfs mount tank/home
	rm -rf $ZFSDIR/openssh-4.2p1 $ZFSDIR/postmark-run $ZFSDIR/local $ZFSDIR/mysql-5.1.30
	rm -f $BDIR/$NAME-*.time
}

# Wait for zfs operations to finish
wait_zfs()
{
	sleep 1
	while ps auxw | grep -q '[z][f][s] destroy'
	do
		echo Waiting for zfs destroy to finish
		sleep 1
	done
	echo All clear
}

# Begin a transaction, setting $D, $TFS and $DIR
begin()
{
	SUBNAME=$1
	date | tr -d \\n
	echo ":  starting $NAME $SUBNAME"
	# Give a change to the previous zfs destroys to finish
	wait_zfs
	/usr/bin/time $TIME_OPTS 2>$BDIR/$NAME-daemon-$SUBNAME.time nice -n -10 ../transactiond &
	sleep 1
	if ! RESULT=`$BDIR/../transaction -b $FS`
	then
		echo "Transaction begin on $FS failed" 1>&2
		exit 1
	fi
	set -- $RESULT
	ID=$1
	TFS=$2
	ODIR=$ZFSDIR
	if [ $NAME = transaction ]
	then
		DIR=/Volumes/$TFS/dds
	else
		DIR=$ODIR
	fi
}

abort()
{
	$BDIR/../transaction -a $ID
	$BDIR/../transaction -x
}

# Perform a commit and check that the directory passed is the same
# as the destination dir
commit_check()
{
	if [ "$1" != "" ]
	then
		rm -rf /tmp/$1
		# Avoid pipe, because it drops events
		tar -C $DIR -cf /tmp/$1.tar $1
	fi
	$BDIR/../transaction -c $ID
	if [ "$1" != "" ]
	then
		tar -C /tmp -xf /tmp/$1.tar
		diff -r /tmp/$1 $ODIR/$1 >$BDIR/$SUBNAME.diff
		rm -rf /tmp/$1 /tmp/$1.tar
	fi
	$BDIR/../transaction -x
}

commit()
{
	$BDIR/../transaction -c $ID
	ps -axww -o command,vsz | grep '[t]ransact' >$BDIR/$NAME-daemon-$SUBNAME.vsz
	$BDIR/../transaction -x
}


# Plain
bench_plain()
{
	DIR=$ZFSDIR
	run_prepare plain
	run_mysql_unpack
	run_mysql_config
	run_mysql_make
	run_mysql_test
	run_mysql_install
	run_ssh_unpack
	run_ssh_config
	run_ssh_make
	#run_trashdir
	run_postmark
}

# While a transaction is in progress
bench_transaction()
{
	run_prepare $1

	begin mysql-unpack
	run_mysql_unpack
	commit mysql-5.1.30

	begin mysql-config
	run_mysql_config
	commit mysql-5.1.30

	begin mysql-make
	run_mysql_make
	commit mysql-5.1.30

	begin mysql-test
	run_mysql_test
	commit mysql-5.1.30

	begin mysql-install
	run_mysql_install
	commit mysql-5.1.30

	begin ssh-unpack
	run_ssh_unpack
	commit openssh-4.2p1

	begin ssh-config
	run_ssh_config
	commit openssh-4.2p1

	begin ssh-make
	run_ssh_make
	commit openssh-4.2p1

	#SUBNAME=trashdir begin
	#run_trashdir
	#commit

	begin postmark
	run_postmark
	commit
}

if ! [ -d $ZFSDIR ]
then
	echo Missing directory
	exit 1
fi

cd $BDIR
pmset sleep 0
#bench_transaction transaction
#bench_transaction autocommit
bench_plain
pmset sleep 30
