#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/fsevents.h>
#include <sys/ioctl.h>

#include "transaction.h"
#include "fsevent.h"

#define DEV_FSEVENTS     "/dev/fsevents" // the fsevents pseudo-device
#define EVENT_QUEUE_SIZE 4096            // limited by MAX_KFS_EVENTS

/*
 * Start monitoring FS events.
 */
int
fsevent_start_monitoring()
{
	static int fd = -1;
	struct fsevent_clone_args fca;
	int8_t event_list[FSE_MAX_EVENTS];
	int ret;
	int flags;
	char buffer[65536];

	if (fd == -1)
		if ((fd = open(DEV_FSEVENTS, O_RDONLY)) < 0) {
			syslog(LOG_ERR, "open " DEV_FSEVENTS ": %m");
			return -1;
		}

	// Set to monitor all events
	memset(event_list, FSE_REPORT, sizeof(event_list));
	fca.event_list = event_list;
	fca.num_events = FSE_MAX_EVENTS;
	fca.event_queue_depth = EVENT_QUEUE_SIZE;
	fca.fd = &eventfd;
	if ((ret = ioctl(fd, FSEVENTS_CLONE, (char *)&fca)) < 0) {
		syslog(LOG_ERR, "ioctl(FSEVENTS_CLONE): %m");
		return -1;
	}

	if ((ret = ioctl(eventfd, FSEVENTS_WANT_EXTENDED_INFO, NULL)) < 0) {
		syslog(LOG_ERR, "ioctl(FSEVENTS_WANT_EXTENDED_INFO): %m");
		close(eventfd);
		eventfd = -1;
		return -1;
	}

#define NEED_NONBLOCK
#ifdef NEED_NONBLOCK
	/*
	 * Set nonblocking reads to prevent blocking in order to fill our buffer.
	 * (This seems to be the default).
	 */
	if ((flags = fcntl(eventfd, F_GETFL, 0)) < 0 ||
	    (flags |= O_NONBLOCK, fcntl(eventfd, F_SETFL, flags) < 0)) {
		syslog(LOG_ERR, "error setting nonblocking mode: %m");
		close(eventfd);
		eventfd = -1;
		return -1;
	}
#endif
#ifdef DRAIN_EVENTS
	// Drain queue of past events
	syslog(LOG_DEBUG, "Draining event queue");
        read(eventfd, buffer, sizeof(buffer));
	syslog(LOG_DEBUG, "Event queue drained");
#endif

	#ifdef DEBUG
	syslog(LOG_DEBUG, "fs events logger started on fd %d", eventfd);
	#endif
	return 0;
}


/*
 * Stop monitoring FS events and drain event queue.
 */
int
fsevent_stop_monitoring()
{
	#ifdef DEBUG
	syslog(LOG_DEBUG, "fs events logger on fd %d stopped", eventfd);
	#endif
	close(eventfd);
	eventfd = -1;
}

// Get a value from the event buffer
#define getval(x) do { \
	if (nbytes < sizeof(x)) { \
		transaction_abort_all(0, "short event read"); \
		return; \
	} \
	memcpy(&x, ptr, sizeof(x)); \
	ptr += sizeof(x); \
	nbytes -= sizeof(x); \
} while (0)

/*
 * Read (and maybe process) file systems event that are available for reading.
 */
void
fsevent_read(void)
{
	ssize_t nbytes;
	static char *buffer;

	assert(eventfd != -1);
	if (buffer == NULL) {
		buffer = malloc(EVENT_BUFFER_SIZE);
		if (buffer == NULL) {
			syslog(LOG_ERR, "event buffer allocation: %m");
			transaction_abort_all(0, "out of memory");
			return;
		}
	}
        if ((nbytes = read(eventfd, buffer, EVENT_BUFFER_SIZE)) < 0) {
		syslog(LOG_ERR, "event reading error: %m");
		transaction_abort_all(0, "error reading event");
		return;
	}
	#ifdef DEBUG_DROPPED_EVENTS
	syslog(LOG_DEBUG, "read %d bytes", nbytes);
	#endif
	if (event_queue_empty()) {
		if (nbytes < EVENT_HIGH_WATER_MARK)
			fsevent_process(buffer, nbytes);
		else {
			/* We can't keep up, queue them. */
			event_queue_add(buffer, nbytes);
			buffer = NULL;
		}
	} else {
		event_queue_add(buffer, nbytes);
		buffer = NULL;
		if (nbytes < EVENT_LOW_WATER_MARK)
			event_queue_process_chunk();
	}
}

/*
 * Decode the passed file system events and free the corresponding buffer.
 */
void
fsevent_process(char *buffer, int nbytes)
{
	char *ptr;
	pid_t pid;
	int32_t flags;
	struct fs_event ev;
	ptr = buffer;

	// Process events
	while (nbytes) {
		memset(&ev, 0, sizeof(ev));
		getval(ev.type);
		flags = FSE_GET_FLAGS(ev.type);
		ev.type &= FSE_TYPE_MASK;
		getval(pid);

		if (ev.type == FSE_EVENTS_DROPPED) {
			syslog(LOG_ERR, "FSE_EVENTS_DROPPED nbytes=%d type=0x%x flags=0x%x pid=%d", nbytes, ev.type, flags, pid);
			transaction_abort_all(0, "FSE_EVENTS_DROPPED");
			return;
		}
		if (flags & FSE_CONTAINS_DROPPED_EVENTS) {
			transaction_abort_all(0, "FSE_CONTAINS_DROPPED_EVENTS");
			return;
		}

		// Process arguments
		for (;;) {
			uint16_t argtype, size;

			getval(argtype);
			if (argtype == FSE_ARG_DONE) {
				transaction_fsevent(&ev);
				break;
			}
			getval(size);
			// Establish precondition: argument data is available
			if (nbytes < size) {
				transaction_abort_all(0, "short event arg read");
				return;
			}
			switch (argtype) {
			case FSE_ARG_STRING:		// next arg is null-terminated string
				if (ev.name)
					ev.name2 = ptr;
				else
					ev.name = ptr;
				break;
			case FSE_ARG_INO:		// next arg is the inode number (ino_t)
				if (ev.ino)
					memcpy(&ev.ino2, ptr, sizeof(ev.ino2));
				else
					memcpy(&ev.ino, ptr, sizeof(ev.ino));
				break;
			case FSE_ARG_DEV:		// next arg is the file's dev_t
				if (ev.dev)
					memcpy(&ev.dev2, ptr, sizeof(ev.dev2));
				else
					memcpy(&ev.dev, ptr, sizeof(ev.dev));
				break;
			// Arguments we ignore
			case FSE_ARG_UID:		// next arg is the file's uid (uid_t)
			case FSE_ARG_GID:		// next arg is the file's gid (gid_t)
				break;
			case FSE_ARG_MODE:		// next arg is the file's mode (as an int32, file type only)
				memcpy(&ev.mode, ptr, sizeof(ev.mode));
				break;
			case FSE_ARG_INT32:		// next arg is a 32-bit int
			case FSE_ARG_INT64:		// next arg is a 64-bit int
			case FSE_ARG_RAW:		// next arg is a length followed by a void ptr
				break;

			// Arguments we shouldn't see
			case FSE_ARG_FINFO:		// next arg is a packed finfo (dev, ino, mode, uid, gid) (kernel-side only)
			case FSE_ARG_PATH:		// next arg is a full path (not used?)
			case FSE_ARG_VNODE:		// vnode pointer (kernel-side only)
			default:
				transaction_abort_all(0, "unexpected file system event");
				syslog(LOG_ERR, "unexpected FSE %d", argtype);
				return;
			}
			nbytes -= size;
			ptr += size;
		}
	}
}
