/*
 * Transaction processing logic
 *
 * $Id: logic.cpp,v 1.38 2009/01/12 10:06:08 dds Exp $
 */

#include <vector>
#include <map>
#include <list>
#include <set>
#include <queue>
#include <string>
#include <iostream>
#include <ostream>
#include <sstream>		// ostringstream
#include <iomanip>		// setw


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>		// lock constants
#include <sys/fsevents.h>	// FSE_*

#include <copyfile.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdlib.h>
#include <pwd.h>

using namespace std;

#include "util.h"
#include "timestamp.h"
#include "transaction.h"
#include "logic.h"
#include "fsevent.h"

/*
 * Execute the specified command, log any errors, and return the result.
 */
static int
system(const string &cmd, string &err_msg)
{
	int ret = system(cmd.c_str());
	if (ret != 0)
		err_msg = cmd + " failed";
	return (ret);
}

/*
 * Return the user name correspoinding the specified uid
 */
static string
user_name(uid_t uid)
{
	return getpwuid(uid)->pw_name;
}
/*
 * Return the device where the specified file resides.
 * Assume that the name specifies a valid filesystem
 */
static dev_t
get_dev(const char *name)
{
	struct stat sb;

	if (stat(name, &sb) == -1) {
		syslog(LOG_ERR, "stat(%s): %m", name);
		return 0;
	}
	return sb.st_dev;
}

/*
 * Get/release exclusive access to the specified filesystems.
 * These are stub functions for the corresponding operating sytem calls.
 */
static int
fslock(char *fs_array[], int n, int flag)
{
	return 0;
}

/*
 * A queue of raw event data chunks.
 * This gets used when we can't decode events at real time.
 */
class EventQueue {
private:
	/* Queue element */
	class EventData {
	private:
		char *data;
		int nbytes;
	public:
		EventData(char *d, int n) : data(d), nbytes(n) {}
		char * get_data() const { return data; }
		int get_nbytes() const { return nbytes; }
	};
	typedef queue<EventData> EventQueueStore;
	static EventQueueStore q;
	static int total_bytes;
public:
	static void add(char *buffer, int nbytes) {
		realloc(buffer, nbytes);
		q.push(EventData(buffer, nbytes));
		total_bytes += nbytes;
	}
	static bool empty() { return q.empty(); }
	/*
	 * Decode two raw event chunks.
	 * This allows event decoding to catch up with event reading.
	 */
	static void process_chunk() {
		for (int i = 0; i < 2; i++) {
			fsevent_process(q.front().get_data(), q.front().get_nbytes());
			free(q.front().get_data());
			total_bytes -= q.front().get_nbytes();
			q.pop();
		}
	}
	/* Decode all pending raw event chunks. */
	static void process_all() {
		while (!q.empty()) {
			fsevent_process(q.front().get_data(), q.front().get_nbytes());
			q.pop();
		}
		total_bytes = 0;
	}
	/* Free all pending raw event chunks. */
	static bool clear() {
		while (!q.empty()) {
			free(q.front().get_data());
			q.pop();
		}
		total_bytes = 0;
	}
	static int get_number_of_bytes() {
		return total_bytes;
	}
	static int get_number_of_elements() {
		return q.size();
	}
};

EventQueue::EventQueueStore EventQueue::q;
int EventQueue::total_bytes;

int
event_queue_empty()
{
	return EventQueue::empty();
}

void
event_queue_add(char *buffer, int nbytes)
{
	EventQueue::add(buffer, nbytes);
}

void
event_queue_process_chunk()
{
	EventQueue::process_chunk();
}
class TransactedObject {
private:
	Timestamp commit_time;		// Time this object was committed
	string name;				// Required only to create meaningful conflict messages
	typedef map <string, TransactedObject *> LiveObjectsStore;
	static LiveObjectsStore live_objects;	// All live objects
	static int nobject;
public:
	TransactedObject(const string &n) : commit_time(0), name(n) {
		nobject++;
	}
	// Return a transacted object (new or existing) for the specified name
	static TransactedObject *get(const string &name);
	// List the live transacted objects
	static void list(int fd);
	static void clear_store() {
		for (LiveObjectsStore::const_iterator i = live_objects.begin(); i != live_objects.end(); i++)
			delete i->second;
		live_objects.clear();
	}
	Timestamp get_commit_time() const {
		return commit_time;
	}
	string get_name() const {
		return name;
	}
	void set_commit_time(Timestamp t) {
		commit_time = t;
	}
	static int get_nobjects() {
		return nobject;
	}
};

int TransactedObject::nobject;
// All live objects
TransactedObject::LiveObjectsStore TransactedObject::live_objects;

void
TransactedObject::list(int fd)
{
	ostringstream line;
	line << setw(11) << right << "Commit Time" << "  Name" << endl;
	for (LiveObjectsStore::const_iterator i = live_objects.begin(); i != live_objects.end(); i++)
		line << setw(11) << right << i->second->commit_time << "  " << i->first << endl;
	int len = line.str().length();
	writen(fd, &len, sizeof(len));
	writen(fd, line.str().c_str(), len);
}

class Transaction;

class MonitoredFilesystem {
private:
	typedef map<dev_t, MonitoredFilesystem> FilesystemStore;
	// Associated transaction; NULL if it is on the main (auto-committed) transaction path
	Transaction *tr;
	string mountpoint;	// Where this filesystem is mounted
	string other_mountpoint;// The mountpoint of the corresponding other (trunk/transaction) fs
	int nref;		// Number of references to this fs
	// Filesystems we monitor
	static FilesystemStore monitored;
public:
	MonitoredFilesystem(Transaction *t, const string &mp) :
		tr(t), mountpoint(mp), nref(1) {}
	// Add a new filesystem to be monitored for a corresponding transaction (or NULL for trunk)
	// Set dev_t fs to the device where the filesystem resides.
	// For transaction branches trunk_fs should be the device where the trunk is mounted
	static void add(const string &name, Transaction *t, dev_t &fs) {
		char *mnt = zfs_mountpoint(name.c_str());
		fs = get_dev(mnt);
		FilesystemStore::iterator l2, l = monitored.find(fs);
		#ifdef DEBUG
		syslog(LOG_DEBUG, "Start monitoring %s (%x)", name.c_str(), fs);
		#endif
		if (l == monitored.end())
			monitored.insert(FilesystemStore::value_type(fs, MonitoredFilesystem(t, mnt)));
		else {
			assert(t == NULL);
			assert(l->second.tr == NULL);
			l->second.nref++;
		}
		free(mnt);
	}
	// Set the mountpoint name of a transacted fs for the trunk fs and vice-versa
	static void set_other_mountpoint(dev_t dev, const string &name) {
		FilesystemStore::iterator l = monitored.find(dev);
		assert (l != monitored.end());
		l->second.other_mountpoint = name;
	}
	// Return the mountpoint name of a filesystem
	static string get_mountpoint(dev_t dev) {
		FilesystemStore::const_iterator l = monitored.find(dev);
		assert (l != monitored.end());
		return l->second.mountpoint;
	}
	// Return the name of a transacted fs file from the name of the trunk fs and vice-versa
	static string get_other_name(dev_t dev, /* copy in */ string name) {
		FilesystemStore::const_iterator l = monitored.find(dev);
		assert (l != monitored.end());
		const MonitoredFilesystem &fs(l->second);
		name.erase(0, fs. mountpoint.length());
		name.insert(0, fs.other_mountpoint);
		return name;
	}
	// Stop monitoring the specified filesystem
	static void remove(dev_t f) {
		FilesystemStore::iterator l = monitored.find(f);
		assert (l != monitored.end());
		l->second.nref--;
		if (l->second.nref == 0)
			monitored.erase(l);
	}
	enum Monitoring {	// How we monitor events from a fileystem
		fm_ignore,	// Ignore events
		fm_trunk,	// Events belong to the trunk, autocommit them
		fm_trans	// Events belong to a transaction, add them
	};
	// Return the type of monitoring to perform, setting t to point to the transaction for fm_trans
	static enum Monitoring type(dev_t fs, Transaction * &t) {
		FilesystemStore::const_iterator l = monitored.find(fs);
		if (l == monitored.end())
			return fm_ignore;
		else if (l->second.tr) {
			t = l->second.tr;
			return fm_trans;
		} else
			return fm_trunk;
	}
	// List the monitored filesystems
	static void list(int fd);
};

// Filesystems we monitor
MonitoredFilesystem::FilesystemStore MonitoredFilesystem::monitored;

// An event within a transaction
class Event {
public:
	// Must be list to avoid invalidating iterators when we add elements
	typedef list<Event> CompressedEventStore;
private:
	int type;		// Type of event (e.g. create, unlink, rename)
	string name, name2;	// File's name
	dev_t dev;		// Associated device
	int mode;		// File's mode
	static int nevent;	// Number of events created
	CompressedEventStore::const_iterator compressed_start;
public:
	Event(struct fs_event *e) :
	    type(e->type), name(e->name), dev(e->dev), mode(e->mode) {
		assert(e->dev2 == 0 || e->dev2 == e->dev);
		if (e->name2)
			name2 = e->name2;
		nevent++;
	}
	Event() : type(FSE_INVALID) {}
	int get_type() const { return type; }
	string get_name() const { return name; }
	string get_name2() const { return name2; }
	dev_t get_dev() const { return dev; }
	CompressedEventStore::const_iterator get_compressed_start() const { return compressed_start; }
	void set_compressed_start(CompressedEventStore::const_iterator i) { compressed_start = i; }
	bool is_dir() const { return S_ISDIR(mode); }
	static int get_nevents() { return nevent; }
};

int Event::nevent;

// A user-level transaction
class Transaction {
private:
	typedef vector<Event> EventStore;
	typedef map<int, Transaction *> TransactionStore;
	typedef set<TransactedObject *> TransactedObjectStore;	// TODO: could use shared_ptr here
								// See http://www.boost.org/doc/libs/1_37_0/libs/smart_ptr/shared_ptr.htm
	int id;			// Transaction id
	int errfd;		// File descriptor for sending error messages
	static int maxid;	// Largest id assigned
	vector<string> fs;	// Filesystems under control (transacted only, not trunks)
	char **fs_array;	// As an array
	vector<bool> clone_created;	// Created clones to destroy when done
	vector<bool> snap_created;	// Created snapshots to destroy when done
	uid_t uid;		// Id of user controlling the transaction
	Timestamp start_time;	// Timestamp indicating relative start time
	TransactedObjectStore modified_objects;	// Objects this transaction has modified
	set <dev_t> assoc_fs;	// All associated filesystems
	EventStore events;	// Events associated with this transaction
	Event::CompressedEventStore compressed_events;	// Rename and delete events associated with this transaction
	static TransactionStore store;
	// Internal utility methods
	void replay_events();
	static string other_name(dev_t dev, const string &name) {
		return MonitoredFilesystem::get_other_name(dev, name);
	}
public:
	Transaction(uid_t xuid, int xerrfd) : id(++maxid), uid(xuid), errfd(xerrfd), fs_array(NULL) {}
	~Transaction();
	// Return a  pointer to the transaction with the specified id of the specified user
	static Transaction *get(int trid, uid_t uid, string &err_msg);
	// List the ongoing transactions
	static void list(int fd);
	static void all_events_list(int fd);
	void events_list(int fd);
	// Abort all, or a user's transactions
	static void abort_all(uid_t uid, const char *msg);
	// Return the number of active transactions
	static TransactionStore::size_type get_active_tr_n() {
		return store.size();
	}
	// Reset the id counter
	static void id_reset() { maxid = 0; }
	// Add a modified object for this transaction
	void add_modified_object(TransactedObject *t) {
		modified_objects.insert(t);
	}
	// Forced abort: abort displayin the specified message
	void abort_msg(const string &msg);
	// Return number of filesystems
	size_t get_nfs() const {
		return fs.size();
	}
	// Return the last known name with which a file is known
	string last_known_name(EventStore::const_iterator start, const string &name);
	// Set the name of the given filesystem
	void add_fs(const string &name) {
		fs.push_back(name);
		clone_created.push_back(false);
		snap_created.push_back(false);
	}
	string get_snapshot_name(int n) const {
		ostringstream name;
		name << fs[n] << "@tr-sn" << id;
		return name.str();
	}
	string get_clone_name(int n) const {
		ostringstream name;
		name << fs[n] << TR_SUFFIX << id;
		return name.str();
	}
	Timestamp get_start_time() const {
		return start_time;
	}
	uid_t get_uid() const { 
		return uid;
	}
	void set_start_time(Timestamp t) {
		start_time = t;
	}
	int get_id() const {
		return id;
	}
	int begin(string &err_msg);
	int commit(string &err_msg);
	// Remove clone and snapshot
	void cleanup();
	void add_event(const Event &e) {
		events.push_back(e);
	}
	inline friend bool operator <(const Transaction &a, const Transaction &b);
};

int Transaction::maxid;
Transaction::TransactionStore Transaction::store;

inline bool
operator <(const Transaction &a, const Transaction &b)
{
	return (a.id < b.id);
}

// List the ongoing transactions
void
Transaction::list(int fd)
{
	ostringstream line;
	line << setw(3) << right << "Id" << "  " << setw(20) << left << "Filesystem" << setw(9) << left << "User" << setw(9) << right << "Created" << setw(9) << right << "Objects" << endl;
	for (TransactionStore::const_iterator i = store.begin(); i != store.end(); i++) {
		Transaction &t = *(i->second);
		for (vector<string>::const_iterator j = t.fs.begin(); j != t.fs.end(); j++)
			line << setw(3) << right << t.id << "  " << setw(20) << left << *j << setw(9) << left << user_name(t.uid) << setw(9) << right << t.start_time << setw(9) << right << t.modified_objects.size() << endl;
	}
	int len = line.str().length();
	writen(fd, &len, sizeof(len));
	writen(fd, line.str().c_str(), len);
}

// List recorded events for all transactions
void
Transaction::all_events_list(int fd)
{
	for (TransactionStore::const_iterator i = store.begin(); i != store.end(); i++)
		i->second->events_list(fd);
}

// List recorded events
void
Transaction::events_list(int fd)
{
	ostringstream line;
	line << setw(3) << right << "Id" << setw(20) << right << "Event" << " " << "File(s)" << endl;
	for (EventStore::const_iterator i = events.begin(); i != events.end(); i++) {
		const Event &e(*i);
		line << setw(3) << right << id;
		switch (e.get_type()) {
		case FSE_CREATE_FILE:
			line << setw(20) << "CREATE_FILE" << ' ' << e.get_name() << endl;
			break;
		case FSE_CONTENT_MODIFIED:
			line << setw(20) << "CONTENT_MODIFIED" << ' ' << e.get_name() << endl;
			break;
		case FSE_CREATE_DIR:
			line << setw(20) << "CREATE_DIR" << ' ' << e.get_name() << endl;
			break;
		case FSE_STAT_CHANGED:
			line << setw(20) << "STAT_CHANGED" << ' ' << e.get_name() << endl;
			break;
		case FSE_CHOWN:
			line << setw(20) << "CHOWN" << ' ' << e.get_name() << endl;
			break;
		case FSE_FINDER_INFO_CHANGED:
			line << setw(20) << "FINDER_INFO_CHANGED" << ' ' << e.get_name() << endl;
			break;
		case FSE_XATTR_MODIFIED:
			line << setw(20) << "XATTR_MODIFIED" << ' ' << e.get_name() << endl;
			break;
		case FSE_XATTR_REMOVED:
			line << setw(20) << "XATTR_REMOVED" << ' ' << e.get_name() << endl;
			break;
		case FSE_RENAME:
			line << setw(20) << "RENAME" << ' ' << e.get_name() << ' ' << e.get_name2() << endl;
			break;
		case FSE_DELETE:
			line << setw(20) << "DELETE" << ' ' << e.get_name() << endl;
			break;
		case FSE_EXCHANGE:
			line << setw(20) << "EXCHANGE" << ' ' << e.get_name() << ' ' << e.get_name2() << endl;
			break;
		}
	}
	int len = line.str().length();
	writen(fd, &len, sizeof(len));
	writen(fd, line.str().c_str(), len);
}

// Abort all, or a user's transactions
void
Transaction::abort_all(uid_t uid, const char *msg)
{
	vector <Transaction *> to_abort;

	syslog(LOG_INFO, "aborting all transactions: %s", msg);
	/* Do it in two steps, because delete changes the underlying store. */
	for (TransactionStore::iterator i = store.begin(); i != store.end(); i++)
		if (uid == 0 || i->second->uid == uid)
			to_abort.push_back(i->second);
	for (vector <Transaction *>::const_iterator i = to_abort.begin(); i != to_abort.end(); i++) {
		(*i)->abort_msg(msg);
		delete(*i);
	}
}

int
Transaction::begin(string &err_msg)
{
	/* Convert vector to array of C strings. */
	fs_array = (char **)calloc(sizeof(char *), get_nfs());
	if (fs_array == NULL) {
		err_msg = "out of memory";
		return -1;
	}
	for (int i = 0; i < get_nfs(); i++)
		if ((fs_array[i] = strdup(fs[i].c_str())) == NULL) {
			err_msg = "out of memory";
			return -1;
		}
	if (fslock(fs_array, get_nfs(), LOCK_EX) < 0) {
		err_msg = "fslock failed";
		return -1;
	}
	/* Snapshot and clone */
	for (int i = 0; i < get_nfs(); i++) {
		ostringstream cmd;
		cmd << "zfs snapshot " << get_snapshot_name(i);
		if (system(cmd.str(), err_msg) == 0)
			snap_created[i] = true;
		else
			return -1;
		cmd.str("");
		cmd << "zfs clone " << get_snapshot_name(i) << " " << get_clone_name(i);
		if (system(cmd.str(), err_msg) == 0)
			clone_created[i] = true;
		else
			return -1;
	}
	set_start_time(Timestamp());
	store.insert(TransactionStore::value_type(id, this));
	for (int i = 0; i < get_nfs(); i++) {
		dev_t tdev, dev;
		MonitoredFilesystem::add(fs[i], NULL, tdev);
		MonitoredFilesystem::add(get_clone_name(i), this, dev);
		// Mark them for eventual removal
		assoc_fs.insert(tdev);
		assoc_fs.insert(dev);
		// Establish cross-links
		MonitoredFilesystem::set_other_mountpoint(tdev, MonitoredFilesystem::get_mountpoint(dev));
		MonitoredFilesystem::set_other_mountpoint(dev, MonitoredFilesystem::get_mountpoint(tdev));
	}
	if (store.size() == 1)
		fsevent_start_monitoring();
	fslock(fs_array, get_nfs(), LOCK_UN);
	if (eventfd == -1) {
		err_msg = "unable to start event monitoring";
		return -1;
	}
	return id;
}

class EventQueue;

int
Transaction::commit(string &err_msg)
{
	if (fslock(fs_array, get_nfs(), LOCK_EX) < 0) {
		err_msg = "fslock failed";
		return -1;
	}
	fsevent_stop_monitoring();
	EventQueue::process_all();
	Timestamp cts;
	// Check for conflicts
	bool to_abort = false;
	string conflicts;
	for (TransactedObjectStore::const_iterator i = modified_objects.begin(); i != modified_objects.end(); i++)
		if ((*i)->get_commit_time() > get_start_time()) {
			conflicts.append((*i)->get_name() + " modified outside the transaction.");
			to_abort = true;
		}
	if (to_abort) {
		err_msg.append("transaction aborted due to conflicts.\n");
		err_msg.append(conflicts);
		return -1;
	}
	// Update commit times
	for (TransactedObjectStore::iterator i = modified_objects.begin(); i != modified_objects.end(); i++)
		(*i)->set_commit_time(cts);
	/// Replay events on main tree
	replay_events();
	fslock(fs_array, get_nfs(), LOCK_UN);
	return 0;
}

/*
 * Return a pointer to the transaction with the specified id
 * and verify that the specified user can access it (or that the
 * user is root).
 * On error log it and return NULL.
 */
Transaction *
Transaction::get(int trid, uid_t uid, string &err_msg)
{
	TransactionStore::iterator i = store.find(trid);
	if (i == store.end()) {
		err_msg = "invalid transaction id";
		return NULL;
	} else {
		Transaction *t = i->second;
		if (t->get_uid() != uid && uid != 0) {
			err_msg = "not transaction owner";
			return NULL;
		}
		return t;
	}
}

// Display the specified abort message
void
Transaction::abort_msg(const string &msg)
{
	ostringstream all;
	all << endl << "\aTransaction " << id << " forced to abort: " << msg << endl;
	write(errfd, all.str().c_str(), all.str().length());
}

Transaction::~Transaction()
{
	if (fslock(fs_array, get_nfs(), LOCK_EX) < 0)
		syslog(LOG_ERR, "fslock failed: %m");
	fsevent_stop_monitoring();
	EventQueue::clear();
	int nfs = get_nfs();
	// Remove it from the store of transactions
	store.erase(id);
	// Adjust sytem's operation
	if (store.size() == 0)
		TransactedObject::clear_store();	// Garbage collection
	// Destroy snapshots and clones that were created
	for (int i = 0; i < nfs; i++) {
		ostringstream cmd;
		string err_msg;
		// Clones must be destroyed before the parent snapshot can be destroyed.
		cmd << "(";
		if (snap_created[i])
			cmd << "zfs destroy " << get_clone_name(i);
		if (snap_created[i] && clone_created[i])
			cmd << " ;";
		if (clone_created[i])
			cmd << "zfs destroy " << get_snapshot_name(i);
		if (snap_created[i] || clone_created[i]) {
			cmd << ") &";
			system(cmd.str(), err_msg);
		}
		#ifdef DEBUG
		syslog(LOG_DEBUG, cmd.str().c_str());
		#endif
		if (err_msg.length())
			syslog(LOG_ERR, err_msg.c_str());
		cmd.str("");
	}
	// Disassociate monitored filesystems
	for (set<dev_t>::const_iterator i = assoc_fs.begin(); i != assoc_fs.end(); i++)
		MonitoredFilesystem::remove(*i);
	if (close(errfd) == -1)
		syslog(LOG_ERR, "close(errfd): %m");
	// Start again monitoring, if needed
	if (store.size() != 0)
		fsevent_start_monitoring();
	fslock(fs_array, nfs, LOCK_UN);
	// Free filesystem array
	if (fs_array) {
		for (int i = 0; i < get_nfs(); i++)
			if (fs_array[i])
				free(fs_array[i]);
		free(fs_array);
	}
}

void
MonitoredFilesystem::list(int fd)
{
	ostringstream line;
	line << setw(4) << right << "#Ref" << "  " << setw(5) << right << "Trans" << "  Filesystem" << " Mounted" << endl;
	for (FilesystemStore::const_iterator i = monitored.begin(); i != monitored.end(); i++) {
		line << setw(4) << right << i->second.nref << "  " << setw(5) << right;
		if (i->second.tr)
			line << i->second.tr->get_id();
		else
			line << '-';
		line << "  0x" << hex << i->first << dec << ' ' << i->second.mountpoint << endl;
	}
	int len = line.str().length();
	writen(fd, &len, sizeof(len));
	writen(fd, line.str().c_str(), len);
}

// Return a transacted object (new or existing) for the specified name
TransactedObject *
TransactedObject::get(const string &trunk_name)
{
	TransactedObject *t;
	LiveObjectsStore::const_iterator i = live_objects.find(trunk_name);
	if (i == live_objects.end()) {
		t = new TransactedObject(trunk_name);
		live_objects.insert(LiveObjectsStore::value_type(trunk_name, t));
	} else
		t = i->second;
	return t;
}


// Report an exit code and (for -ve codes) a message to the calling process.
static void
done(int fd, int code, const string &msg = "")
{
	writen(fd, &code, sizeof(code));
	if (code < 0) {
		int len = msg.length();
		writen(fd, &len, sizeof(len));
		writen(fd, msg.c_str(), len);
	}
}

/*
 * Begin a new transaction.
 * Read from fd a sequence of len, char[] strings specifying the filesystems
 * Write the transaction id, or -1 on failure.
 */
void
transaction_begin(int fd, uid_t uid, int nfs, int errfd)
{
	int len;
	char name[PATH_MAX + 1];
	Transaction *tr = new Transaction(uid, errfd);

	/* Read number of filesystems. */
	for (int i = 0; i < nfs; i++) {
		if (readn(fd, &len, sizeof(len)) < 0 ||
		    readn(fd, name, len) < 0) {
			syslog(LOG_ERR, "failed to read name of filesystem %d", i);
			delete tr;
			return;
		}
		tr->add_fs(string(name, len));
	}
	string err_msg;
	int ret = tr->begin(err_msg);
	if (ret < 0)
		delete tr;
	done(fd, ret, err_msg);
}

void
transaction_abort(int fd, uid_t uid, int trid)
{
	string err_msg;
	Transaction *t = Transaction::get(trid, uid, err_msg);
	if (t == NULL) {
		done(fd, -1, err_msg);
		return;
	}
	delete t;
	done(fd, 0);
}

void
transaction_commit(int fd, uid_t uid, int trid)
{
	string err_msg;
	Transaction *t = Transaction::get(trid, uid, err_msg);
	if (t == NULL) {
		done(fd, -1, err_msg);
		return;
	}
	int ret = t->commit(err_msg);
	delete t;
	done(fd, ret, err_msg);
}

void
transaction_list(int fd, uid_t uid, int ltype)
{
	switch (ltype) {
	case 'e':		// Events
		Transaction::all_events_list(fd);
		break;
	case 'f':		// Filesystems
		MonitoredFilesystem::list(fd);
		break;
	case 'o':		// Transacted objects
		TransactedObject::list(fd);
		break;
	case 't':		// Transactions
		Transaction::list(fd);
		break;
	case 's':		// Statistics
		{
		ostringstream line;
		line << "events: " << Event::get_nevents() <<
			" objects: " << TransactedObject::get_nobjects() <<
			" raw elements: " << EventQueue::get_number_of_elements() <<
			" raw bytes: " << EventQueue::get_number_of_bytes() <<
			endl;
		int len = line.str().length();
		writen(fd, &len, sizeof(len));
		writen(fd, line.str().c_str(), len);
		}
		break;
	}
}

/*
 * Process a file system event that may affect transactions
 */
void
transaction_fsevent(struct fs_event *ev)
{
	Transaction *t;
	switch (MonitoredFilesystem::type(ev->dev, /* return value */ t)) {
	case MonitoredFilesystem::fm_ignore:	// Ignore events not associated with fs we monitor
		break;
	case MonitoredFilesystem::fm_trunk:	// Events belong to the trunk, autocommit them
		TransactedObject::get(ev->name)->set_commit_time(Timestamp());
		break;
	case MonitoredFilesystem::fm_trans:	// Events belong to a transaction, add them
		string trunk_name(MonitoredFilesystem::get_other_name(ev->dev, ev->name));
		t->add_modified_object(TransactedObject::get(trunk_name));
		t->add_event(ev);
		break;
	}
	if (ev->name2) {
		Transaction *t2;
		dev_t dev2 = ev->dev2 ? ev->dev2 : ev->dev;
		switch (MonitoredFilesystem::type(dev2, /* return value */ t2)) {
		case MonitoredFilesystem::fm_ignore:	// Ignore events not associated with fs we monitor
			break;
		case MonitoredFilesystem::fm_trunk:	// Events belong to the trunk, autocommit them
			TransactedObject::get(ev->name2)->set_commit_time(Timestamp());
			break;
		case MonitoredFilesystem::fm_trans:	// Events belong to a transaction, add them
			if (t2 != t) {
				/*
				 * Events spanning two transactions are not allowed.
				 * Until a transaction is committed, nobody is allowed to see
				 * beyond its event horizon.
				 * Normally, this should not happen, because rename(2) and
				 * exchangedata(2) would return EXDEV.
				 */
				t->abort_msg("event affects two active transactions");
				delete t;
				t2->abort_msg("event affects two active transactions");
				delete t2;
				return;
			}
			string trunk_name2(MonitoredFilesystem::get_other_name(dev2, ev->name2));
			t2->add_modified_object(TransactedObject::get(trunk_name2));
			break;
		}
	}
}

/*
 * An internal error or user command causes us to abort all transactions
 */
void
transaction_abort_all(uid_t uid, const char *msg)
{
	#ifdef DEBUG
	syslog(LOG_DEBUG, "aborting all transactions: %s", msg);
	#endif
	#ifdef NO_DAEMON
	syslog(LOG_INFO, "Processed %d events and %d objects", Event::get_nevents(), TransactedObject::get_nobjects());
	#endif
	Transaction::abort_all(uid, msg);
	// Reset counters to simplify regression testing
	if (uid == 0) {
		Timestamp::reset();
		Transaction::id_reset();
	}
}

/*
 * Return the last name for which name is known, after renames and deletes,
 * starting at event start.
 * Return the empty string if name was in the end deleted.
 */
string
Transaction::last_known_name(EventStore::const_iterator start, const string &name)
{
	bool done = false;
	bool waiting_rename = false;
	string result(name);

	for (Event::CompressedEventStore::const_iterator i = start->get_compressed_start(); i != compressed_events.end(); i++) {
		const Event &e(*i);

		switch (e.get_type()) {
		case FSE_DELETE:
			if (e.get_name() == result)
				return "";
			break;
		case FSE_RENAME:
			if (e.get_name() == result) 		// Exact match
				result = e.get_name2();
			else if (result.length() > e.get_name().length() &&	// Path match
			    e.get_name() == result.substr(0, e.get_name().length()) &&
			    result[e.get_name().length()] == '/')
				result.replace(0, e.get_name().length(), e.get_name2());
			break;
		case FSE_EXCHANGE:		// exchangedata(2)
			if (e.get_name() == result)		// Exact match
				result = e.get_name2();
			else if (e.get_name2() == result)	// Exact match
				result = e.get_name();
			break;
		}
	}
	return result;
}

// Replay events on main tree
void
Transaction::replay_events()
{

	// Now replay the events on the trunk filesystem
	int old_umask = umask(0);
	if (seteuid(uid) == -1)
		syslog(LOG_ERR, "seteuid(%d): %m", uid);

	// Compress events that affect names to avoid quadratic behavior of last_known_name
	compressed_events.push_back(Event());	// NOP, placeholder
	for (EventStore::iterator i = events.begin(); i != events.end(); i++) {
		Event &e(*i);
		e.set_compressed_start(--compressed_events.end());
		switch (e.get_type()) {
		case FSE_DELETE:
		case FSE_RENAME:
		case FSE_EXCHANGE:
			compressed_events.back() = e;
			compressed_events.push_back(Event());	// NOP, placeholder
		}
	}
	/*
	 * We execute the event log playback with the user's privileges.
	 * Because we're playing back events that have already succeeded, and
	 * there are no conflicting events, we assume that the user has sufficient
	 * privileges to execute these actions.
	 */
	for (EventStore::const_iterator i = events.begin(); i != events.end(); i++) {
		const Event &e(*i);
		string str_src(last_known_name(i, e.get_name()));
		string str_dst(other_name(e.get_dev(), e.get_name()));
		const char *src = str_src.c_str();
		const char *dst = str_dst.c_str();

		switch (e.get_type()) {
		case FSE_CREATE_FILE:
		case FSE_CONTENT_MODIFIED:
			if (str_src.length() == 0)
				continue; // Skip deleted files
			#ifdef DEBUG
			syslog(LOG_DEBUG, "copyall(%s, %s)", src, dst);
			#endif
			copyfile(src, dst, NULL, COPYFILE_ALL);
			break;
		case FSE_CREATE_DIR:
			#ifdef DEBUG
			syslog(LOG_DEBUG, "mkdir(%s)", dst);
			#endif
			mkdir(dst, 0777);
			if (str_src.length() == 0)
				continue; // Skip deleted files
			#ifdef DEBUG
			syslog(LOG_DEBUG, "copymeta(%s, %s)", src, dst);
			#endif
			copyfile(src, dst, NULL, COPYFILE_METADATA);
			break;
		case FSE_STAT_CHANGED:
			if (str_src.length() == 0)
				continue; // Skip deleted files
			#ifdef DEBUG
			syslog(LOG_DEBUG, "copystat(%s, %s)", src, dst);
			#endif
			copyfile(src, dst, NULL, COPYFILE_STAT);
			break;
		case FSE_CHOWN:
			if (str_src.length() == 0)
				continue; // Skip deleted files
			#ifdef DEBUG
			syslog(LOG_DEBUG, "copysec(%s, %s)", src, dst);
			#endif
			copyfile(src, dst, NULL, COPYFILE_SECURITY);
			break;
		case FSE_FINDER_INFO_CHANGED:
		case FSE_XATTR_MODIFIED:
		case FSE_XATTR_REMOVED:
			if (str_src.length() == 0)
				continue; // Skip deleted files
			#ifdef DEBUG
			syslog(LOG_DEBUG, "copyxattr(%s, %s)", src, dst);
			#endif
			copyfile(src, dst, NULL, COPYFILE_XATTR);
			break;
		case FSE_RENAME:
			{
			string str_dst2(other_name(e.get_dev(), e.get_name2()));
			const char *dst2 = str_dst2.c_str();
			#ifdef DEBUG
			syslog(LOG_DEBUG, "rename(%s, %s)", dst, dst2);
			#endif
			rename(dst, dst2);
			break;
			}
		case FSE_DELETE:
			if (e.is_dir()) {
				#ifdef DEBUG
				syslog(LOG_DEBUG, "rmdir(%s)", dst);
				#endif
				rmdir(dst);
			} else {
				#ifdef DEBUG
				syslog(LOG_DEBUG, "unlink(%s)", dst);
				#endif
				unlink(dst);
			}
			break;
		case FSE_EXCHANGE:		// exchangedata(2)
			{
			string str_src2(last_known_name(i, e.get_name()));
			string str_dst2(other_name(e.get_dev(), e.get_name2()));
			const char *src2 = str_src.c_str();
			const char *dst2 = str_dst.c_str();
			if (str_src.length() != 0) {
				#ifdef DEBUG
				syslog(LOG_DEBUG, "copyall(%s, %s)",src, dst2);
				#endif
				copyfile(src, dst2, NULL, COPYFILE_ALL);
			}
			if (str_src2.length() != 0) {
				#ifdef DEBUG
				syslog(LOG_DEBUG, "copyall(%s, %s)",src2, dst);
				#endif
				copyfile(src2, dst, NULL, COPYFILE_ALL);
			}
			break;
			}
		}
	}
	if (seteuid(0) == -1)
		syslog(LOG_ERR, "seteuid(0): %m");
	(void)umask(old_umask);
}
