/*
 * Transaction client.
 */

#include <sys/errno.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "transaction.h"
#include "util.h"

extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;
extern int optreset;

int getopt(int argc, char * const argv[], const char *optstring);

static char *progname;

static void
usage()
{
	fprintf(stderr, "Usage:");
	fprintf(stderr, "\t%s -b filesystem ...\n", progname);
	fprintf(stderr, "\t%s -c id\n", progname);
	fprintf(stderr, "\t%s -a id\n", progname);
	fprintf(stderr, "\t%s -l [e|f|o|t|s]\n", progname);
	fprintf(stderr, "\t%s -x|-A\n", progname);
	exit(1);
}

/* Copy a length-headed string from fd to stream. */
static void
copystring(int fd, FILE *f)
{
	int n;
	char *s;

	readn(fd, &n, sizeof(n));

	if ((s = malloc(n)) == NULL)
		errx(2, "out of memory");
	readn(fd, s, n);
	fwrite(s, n, 1, f);
}

int
main(int argc, char *argv[])
{
	int sockfd;
	struct sockaddr_un servaddr;
	struct s_command cmd = {0};
	int ret, i, n;
	int c;

	progname = argv[0];

	while ((c = getopt(argc, argv, "Aa:bc:l:x")) != EOF)
			switch (c) {
			case 'l':		// List
				if (cmd.cmd ||
				    strlen(optarg) != 1 ||
				    strchr("efots", *optarg) == NULL)
					usage();
				cmd.n = *optarg;
				cmd.cmd = c;
				break;
			case 'x':		// Exit
				if (geteuid() != 0)
					errx(1, "Only root can issue this command");
				/* FALLTHROUGH */
			case 'A':		// Abort all
			case 'b':		// Begin
				if (cmd.cmd)
					usage();
				cmd.cmd = c;
				break;
			case 'a':		// Abort
			case 'c':		// Commit
				cmd.n = atoi(optarg);
				if (cmd.cmd)
					usage();
				cmd.cmd = c;
				break;
			case '?':
				usage();
			}
	if (!cmd.cmd)
		usage();
	argc -= optind;
	argv += optind;
	switch (cmd.cmd) {
	case 'l':		// List
	case 'a':		// Abort
	case 'A':		// Abort all
	case 'x':		// Exit
	case 'c':		// Commit
		if (argc)
			usage();
		break;
	case 'b':		// Begin
		if (argc == 0)
			usage();
		/* Ensure that the specified filesystems are of ZFS type. */
		for (i = 0; i < argc; i++)
			if (zfs_mountpoint(argv[i]) == NULL)
				exit(1);
		cmd.n = argc;
		break;
	}

	if ((sockfd = socket(AF_LOCAL, SOCK_STREAM, 0)) == -1)
		err(1, "socket");
	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sun_family = AF_LOCAL;
	strcpy(servaddr.sun_path, CMD_FILE);
	if (connect(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
		err(1, "connect");
	writen(sockfd, &cmd, sizeof(cmd));
	switch(cmd.cmd) {
	case 'b':
		/* Send error fd. */
		send_fd(sockfd, fileno(stderr));
		/* Send filesystems. */
		for (i = 0; i < argc; i++) {
			int n = strlen(argv[i]);
			writen(sockfd, &n, sizeof(n));
			writen(sockfd, argv[i], n);
		}
		/* Get return value. */
		readn(sockfd, &ret, sizeof(ret));
		/* Process result. */
		if (ret == -1) {
			fprintf(stderr, "Transaction begin failed: ");
			copystring(sockfd, stderr);
			putc('\n', stderr);
			exit(2);
		} else {
			printf("%d\n", ret);
			for (i = 0; i < argc; i++)
				printf("%s" TR_SUFFIX "%d\n", argv[i], ret);
			exit(0);
		}
	case 'a':		// Abort
	case 'c':		// Commit
		/* Get return value. */
		readn(sockfd, &ret, sizeof(ret));
		/* Process result. */
		if (ret == -1) {
			fprintf(stderr, "Transaction commit or abort failed: ");
			copystring(sockfd, stderr);
			putc('\n', stderr);
			exit(2);
		} else
			exit(0);
	case 'l':
			copystring(sockfd, stdout);
			exit(0);
	case 'A':
	case 'x':
			exit(0);
	}
	/* NOTREACHED */
}
