#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/socket.h>

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "util.h"

/*
 * Guaranteed read for the specified bytes.
 * On short read or error, log the problem and return -1
 * On success return 0.
 */
int
readn(int fd, void *buf, size_t nbyte)
{
	int nread;
	int remain = nbyte;
	char *ptr = buf;

	while (remain > 0) {
		if ((nread = read(fd, ptr, remain)) < 0)
			switch (errno) {
			case EAGAIN:
				usleep(5000);		/* 5ms */
				/* FALLTHROUGH */
			case EINTR:
				nread = 0;
				break;
			default:
				err("read");
				return (-1);
			}
		else if (nread == 0) {
			/* EOF */
			errx("short read");
			return (-1);
		}
		remain -= nread;
		ptr += nread;
	}
	return (0);
}

/*
 * Guaranteed write for the specified bytes.
 * On error, log the problem and return -1
 * On success return 0.
 */
int
writen(int fd, const void *buf, size_t nbyte)
{
	int nwrite;
	int remain = nbyte;
	const char *ptr = buf;

	while (remain > 0) {
		if ((nwrite = write(fd, ptr, remain)) < 0)
			switch (errno) {
			case EAGAIN:
				usleep(5000);		/* 5ms */
				/* FALLTHROUGH */
			case EINTR:
				nwrite = 0;
				break;
			default:
				err("write");
				return (-1);
			}
		remain -= nwrite;
		ptr += nwrite;
	}
	return (0);
}

union fdmsg {
	struct cmsghdr h;
	char buf[CMSG_SPACE(sizeof(int))];
};

/* Receive a file descriptor from a socket. */
int
receive_fd(int sock)
{
	int count;
	char ping;
	struct iovec iov[1];
	struct msghdr msg;
	union fdmsg cmsg;
	struct cmsghdr* h;

	iov[0].iov_base = &ping;
	iov[0].iov_len  = 1;

	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_control = cmsg.buf;
	msg.msg_controllen = sizeof(union fdmsg);
	msg.msg_flags = 0;

	h = CMSG_FIRSTHDR(&msg);
	h->cmsg_len = CMSG_LEN(sizeof(int));
	h->cmsg_level = SOL_SOCKET;
	h->cmsg_type = SCM_RIGHTS;
	*((int*)CMSG_DATA(h)) = -1;

	if ((count = recvmsg(sock, &msg, 0)) < 0) {
		err("recvmsg");
		return -1;
	}
	h = CMSG_FIRSTHDR(&msg);
	if (h == NULL ||
            h->cmsg_len != CMSG_LEN(sizeof(int)) ||
	    h->cmsg_level != SOL_SOCKET ||
	    h->cmsg_type != SCM_RIGHTS ) {
		errx("incorrect value in recvmsg");
		return -1;
	}
	return *((int*)CMSG_DATA(h));
}


/* Send the specified file descriptor to a socket. */
int
send_fd(int sock, int fd)
{
	int ping;
	struct iovec iov[1];
	struct msghdr msg;
	union  fdmsg cmsg;
	struct cmsghdr *h;

	iov[0].iov_base = &ping;
	iov[0].iov_len = 1;

	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_control = cmsg.buf;
	msg.msg_controllen = sizeof(union fdmsg);
	msg.msg_flags = 0;

	h = CMSG_FIRSTHDR(&msg);
	h->cmsg_len = CMSG_LEN(sizeof(int));
	h->cmsg_level = SOL_SOCKET;
	h->cmsg_type = SCM_RIGHTS;
	*((int*)CMSG_DATA(h)) = fd;

	if (sendmsg(sock, &msg, 0) < 0) {
		err("sendmsg");
		return -1;
	}
	return 0;
}

/*
 * Return the mount point of the specified ZFS filesystem,
 * or NULL if the argument does not specify a ZFS filesystem.
 * The name is returned in malloced memory.
 */
char *
zfs_mountpoint(const char *dataset)
{
	char *cmd, *ret, *p;
	FILE *f;
	size_t len;

	asprintf(&cmd, "zfs get -H -o value mountpoint %s", dataset);
	if ((f = popen(cmd, "r")) == NULL) {
		err("zfs get mountpoint %s", dataset);
		free(cmd);
		return NULL;
	}
again:
	p = fgetln(f, &len);
#ifdef DEBUG
	/* Running under gdb causes fgetln to return 0 with EINTR. */
	if (len == 0 && errno == EINTR)
		goto again;
#endif
	if (p && len && p[len - 1] == '\n') {
		p[len - 1] = 0;
		ret = strdup(p);
	} else if (ferror(f)) {
		err("zfs mountpoint failure");
		ret = NULL;
	} else {
		errx("zfs mountpoint failure p=%p len=%d [%s]", p, len, p ? p : "");
		ret = NULL;
	}
	free(cmd);
	pclose(f);
	return ret;
}
